# futures
from __future__ import absolute_import

# standard libraries
import unittest

# third party libraries
# None

# local libraries
# None


class TestTaskClass(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_none(self):
        pass
